@@ -1,15 +1,20 @@
+/* $Id: road_cmd.c 3957 2006-03-18 15:29:24Z Darkvater $ */
+
 #include "stdafx.h"
-#include "ttd.h"
+#include "openttd.h"
+#include "table/sprites.h"
 #include "table/strings.h"
+#include "functions.h"
 #include "map.h"
+#include "tile.h"
 #include "vehicle.h"
 #include "viewport.h"
 #include "command.h"
 #include "player.h"
 #include "town.h"
 #include "gfx.h"
-#include "table/directions.h"
 #include "sound.h"
+#include "depot.h"
 
 /* When true, GetTrackStatus for roads will treat roads under reconstruction
  * as normal roads instead of impassable. This is used when detecting whether
@@ -20,17 +25,17 @@ static bool _road_special_gettrackstatus;
 void RoadVehEnterDepot(Vehicle *v);
 
 
-static bool HasTileRoadAt(uint tile, int i)
+static bool HasTileRoadAt(TileIndex tile, int i)
 {
 	int mask;
 	byte b;
 
-	switch (TileType(tile)) {
+	switch (GetTileType(tile)) {
 	case MP_STREET:
-		b = _map5[tile];
+		b = _m[tile].m5;
 
 		if ((b & 0xF0) == 0) {
-		} else if ((b & 0xF0) == 0x10) {
+		} else if (IsLevelCrossing(tile)) {
 			b = (b&8)?5:10;
 		} else if ((b & 0xF0) == 0x20) {
 			return (~b & 3) == i;
@@ -39,7 +44,7 @@ static bool HasTileRoadAt(uint tile, int i)
 		break;
 
 	case MP_STATION:
-		b = _map5[tile];
+		b = _m[tile].m5;
 		if (!IS_BYTE_INSIDE(b, 0x43, 0x43+8))
 			return false;
 		return ((~(b - 0x43) & 3) == i);
@@ -57,7 +62,7 @@ static bool HasTileRoadAt(uint tile, int i)
 	return HASBIT(b, i);
 }
 
-static bool CheckAllowRemoveRoad(uint tile, uint br, bool *edge_road)
+static bool CheckAllowRemoveRoad(TileIndex tile, uint br, bool *edge_road)
 {
 	int blocks;
 	byte owner;
@@ -76,10 +81,10 @@ static bool CheckAllowRemoveRoad(uint tile, uint br, bool *edge_road)
 		return true;
 
 	// A railway crossing has the road owner in the map3_lo byte.
-	if (IsTileType(tile, MP_STREET) && (_map5[tile] & 0xF0) == 0x10) {
-		owner = _map3_lo[tile];
+	if (IsTileType(tile, MP_STREET) && IsLevelCrossing(tile)) {
+		owner = _m[tile].m3;
 	} else {
-		owner = _map_owner[tile];
+		owner = GetTileOwner(tile);
 	}
 	// Only do the special processing if the road is owned
 	// by a town
@@ -106,7 +111,8 @@ static bool CheckAllowRemoveRoad(uint tile, uint br, bool *edge_road)
 		if (_patches.extra_dynamite)
 			return true;
 
-		t = ClosestTownFromTile(tile, (uint)-1);
+		t = ClosestTownFromTile(tile, _patches.dist_local_authority);
+
 		SetDParam(0, t->index);
 		_error_message = STR_2009_LOCAL_AUTHORITY_REFUSES;
 		return false;
@@ -115,40 +121,53 @@ static bool CheckAllowRemoveRoad(uint tile, uint br, bool *edge_road)
 	return true;
 }
 
-bool IsRoadDepotTile(TileIndex tile)
-{
-	return IsTileType(tile, MP_STREET) &&
-					(_map5[tile] & 0xF0) == 0x20;
-}
-
 uint GetRoadBitsByTile(TileIndex tile)
 {
 	uint32 r = GetTileTrackStatus(tile, TRANSPORT_ROAD);
 	return (byte)(r | (r >> 8));
 }
 
-// cost for removing inner/edge -roads
-static const uint16 _road_remove_cost[2] = {50, 18};
-
-/* Delete a piece of road
- * p1 = piece type
+/** Delete a piece of road.
+ * @param x,y tile coordinates for road construction
+ * @param p1 road piece flags
+ * @param p2 unused
  */
 int32 CmdRemoveRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
+	// cost for removing inner/edge -roads
+	static const uint16 road_remove_cost[2] = {50, 18};
+
 	TileInfo ti;
 	int32 cost;
-	uint tile;
+	TileIndex tile;
+	PlayerID owner;
 	Town *t;
-	/*	true if the roadpiece was always removeable,
-			false if it was a center piece. Affects town ratings drop
-	*/
+	/* true if the roadpiece was always removeable,
+	 * false if it was a center piece. Affects town ratings drop */
 	bool edge_road;
+	byte pieces = (byte)p1;
 
 	SET_EXPENSES_TYPE(EXPENSES_CONSTRUCTION);
 
+	/* Road pieces are max 4 bitset values (NE, NW, SE, SW) */
+	if (pieces >> 4) return CMD_ERROR;
+
 	FindLandscapeHeight(&ti, x, y);
 	tile = ti.tile;
-	t = ClosestTownFromTile(tile, (uint)-1); // needed for town rating penalty
+
+	if (!IsTileType(tile, MP_STREET) && !IsTileType(tile, MP_TUNNELBRIDGE)) return CMD_ERROR;
+
+	// owner for railroad crossing is stored somewhere else
+	// XXX - Fix this so for a given tiletype the owner of the type is in the same variable
+	owner = IsLevelCrossing(tile) ? _m[tile].m3 : GetTileOwner(tile);
+
+	if (owner == OWNER_TOWN && _game_mode != GM_EDITOR) {
+		if (IsTileType(tile, MP_TUNNELBRIDGE)) { // index of town is not saved for bridge (no space)
+			t = ClosestTownFromTile(tile, _patches.dist_local_authority);
+		} else
+			t = GetTown(_m[tile].m2);
+	} else
+		t = NULL;
 
 	// allow deleting road under bridge
 	if (ti.type != MP_TUNNELBRIDGE && !EnsureNoVehicle(tile))
@@ -157,10 +176,9 @@ int32 CmdRemoveRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 	{
 		bool b;
 		_road_special_gettrackstatus = true;
-		b = CheckAllowRemoveRoad(tile, p1, &edge_road);
+		b = CheckAllowRemoveRoad(tile, pieces, &edge_road);
 		_road_special_gettrackstatus = false;
-		if (!b)
-			return CMD_ERROR;
+		if (!b) return CMD_ERROR;
 	}
 
 	if (ti.type == MP_TUNNELBRIDGE) {
@@ -168,23 +186,29 @@ int32 CmdRemoveRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 			return CMD_ERROR;
 
 		if ((ti.map5 & 0xE9) == 0xE8) {
-			if (p1 & 10) goto return_error;
+			if (pieces & 10) goto return_error;
 		} else if ((ti.map5 & 0xE9) == 0xE9) {
-			if (p1 & 5) goto return_error;
+			if (pieces & 5) goto return_error;
 		} else
 			goto return_error;
 
 		cost = _price.remove_road * 2;
 
 		if (flags & DC_EXEC) {
-			_map5[tile] = ti.map5 & 0xC7;
-			_map_owner[tile] = OWNER_NONE;
+			ChangeTownRating(t, -road_remove_cost[(byte)edge_road], RATING_ROAD_MINIMUM);
+			_m[tile].m5 = ti.map5 & 0xC7;
+			SetTileOwner(tile, OWNER_NONE);
 			MarkTileDirtyByTile(tile);
 		}
 		return cost;
 	} else if (ti.type == MP_STREET) {
-		if (!(ti.map5 & 0xF0)) {
-			uint c = p1, t2;
+		// check if you're allowed to remove the street owned by a town
+		// removal allowance depends on difficulty setting
+		if (!CheckforTownRating(tile, flags, t, ROAD_REMOVE)) return CMD_ERROR;
+
+		// XXX - change cascading ifs to switch when doing rewrite
+		if ((ti.map5 & 0xF0) == 0) { // normal road
+			byte c = pieces, t2;
 
 			if (ti.tileh != 0  && (ti.map5 == 5 || ti.map5 == 10)) {
 				c |= (c & 0xC) >> 2;
@@ -201,44 +225,36 @@ int32 CmdRemoveRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 				if (t2&1) cost += _price.remove_road;
 			} while(t2>>=1);
 
-			// check if you're allowed to remove the street owned by a town
-			// removal allowance depends on difficulty setting
-			if(_map_owner[tile] == OWNER_TOWN && _game_mode != GM_EDITOR) {
-				if (!CheckforTownRating(tile, flags, t, ROAD_REMOVE))
-					return CMD_ERROR;
-			}
-
 			if (flags & DC_EXEC) {
-				// checks if the owner is town than decrease town rating by 50 until you have
-				// a "Poor" town rating
-				if(_map_owner[tile] == OWNER_TOWN && _game_mode != GM_EDITOR)
-					ChangeTownRating(t, -_road_remove_cost[(byte)edge_road], RATING_ROAD_MINIMUM);
+				ChangeTownRating(t, -road_remove_cost[(byte)edge_road], RATING_ROAD_MINIMUM);
 
-				_map5[tile] ^= c;
-				if ((_map5[tile]&0xF) == 0)
+				_m[tile].m5 ^= c;
+				if (GB(_m[tile].m5, 0, 4) == 0) {
 					DoClearSquare(tile);
-				else
+				} else {
 					MarkTileDirtyByTile(tile);
-
+				}
 			}
 			return cost;
-		} else if (!(ti.map5 & 0xE0)) {
+		} else if ((ti.map5 & 0xE0) == 0) { // railroad crossing
 			byte c;
 
 			if (!(ti.map5 & 8)) {
 				c = 2;
-				if (p1 & 5) goto return_error;
+				if (pieces & 5) goto return_error;
 			} else {
 				c = 1;
-				if (p1 & 10) goto return_error;
+				if (pieces & 10) goto return_error;
 			}
 
 			cost = _price.remove_road * 2;
 			if (flags & DC_EXEC) {
+				ChangeTownRating(t, -road_remove_cost[(byte)edge_road], RATING_ROAD_MINIMUM);
+
 				ModifyTile(tile,
 					MP_SETTYPE(MP_RAILWAY) |
 					MP_MAP2_CLEAR | MP_MAP3LO | MP_MAP3HI_CLEAR | MP_MAP5,
-					_map3_hi[tile] & 0xF, /* map3_lo */
+					_m[tile].m4 & 0xF, /* map3_lo */
 					c											/* map5 */
 				);
 			}
@@ -304,7 +320,7 @@ static const byte _valid_tileh_slopes_road[3][15] = {
 
 static uint32 CheckRoadSlope(int tileh, byte *pieces, byte existing)
 {
-	if (!(tileh & 0x10)) {
+	if (!IsSteepTileh(tileh)) {
 		byte road_bits = *pieces | existing;
 
 		// no special foundation
@@ -330,29 +346,33 @@ static uint32 CheckRoadSlope(int tileh, byte *pieces, byte existing)
 	return CMD_ERROR;
 }
 
-/* Build a piece of road
- * p1 = piece flags
+/** Build a piece of road.
+ * @param x,y tile coordinates for road construction
+ * @param p1 road piece flags
+ * @param p2 the town that is building the road (0 if not applicable)
  */
-
 int32 CmdBuildRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
 	TileInfo ti;
 	int32 cost;
 	byte pieces = (byte)p1, existing = 0;
-	uint tile;
+	TileIndex tile;
 
 	SET_EXPENSES_TYPE(EXPENSES_CONSTRUCTION);
 
+	/* Road pieces are max 4 bitset values (NE, NW, SE, SW) and town can only be non-zero
+	 * if a non-player is building the road */
+	if ((pieces >> 4) || (_current_player < MAX_PLAYERS && p2 != 0) || !IsTownIndex(p2)) return CMD_ERROR;
+
 	FindLandscapeHeight(&ti, x, y);
 	tile = ti.tile;
 
 	// allow building road under bridge
-	if (ti.type != MP_TUNNELBRIDGE && !EnsureNoVehicle(tile))
-		return CMD_ERROR;
+	if (ti.type != MP_TUNNELBRIDGE && !EnsureNoVehicle(tile)) return CMD_ERROR;
 
 	if (ti.type == MP_STREET) {
 		if (!(ti.map5 & 0xF0)) {
-			if ( ((pieces) & (byte)(ti.map5)) == (pieces))
+			if ((pieces & (byte)ti.map5) == pieces)
 				return_cmd_error(STR_1007_ALREADY_BUILT);
 			existing = ti.map5;
 		} else {
@@ -363,7 +383,7 @@ int32 CmdBuildRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 	} else if (ti.type == MP_RAILWAY) {
 		byte m5;
 
-		if (ti.tileh & 0x10) // very steep tile
+		if (IsSteepTileh(ti.tileh)) // very steep tile
 				return_cmd_error(STR_1000_LAND_SLOPED_IN_WRONG_DIRECTION);
 
 		if(!_valid_tileh_slopes_road[2][ti.tileh]) // prevent certain slopes
@@ -381,9 +401,10 @@ int32 CmdBuildRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 		if (flags & DC_EXEC) {
 			ModifyTile(tile,
 				MP_SETTYPE(MP_STREET) |
-				MP_MAP2_CLEAR | MP_MAP3LO | MP_MAP3HI | MP_MAP5,
+				MP_MAP2 | MP_MAP3LO | MP_MAP3HI | MP_MAP5,
+				p2,
 				_current_player, /* map3_lo */
-				_map3_lo[tile] & 0xF, /* map3_hi */
+				_m[tile].m3 & 0xF, /* map3_hi */
 				m5 /* map5 */
 			);
 		}
@@ -391,7 +412,7 @@ int32 CmdBuildRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 	} else if (ti.type == MP_TUNNELBRIDGE) {
 
 		/* check for flat land */
-		if (ti.tileh & 0x10) // very steep tile
+		if (IsSteepTileh(ti.tileh)) // very steep tile
 				return_cmd_error(STR_1000_LAND_SLOPED_IN_WRONG_DIRECTION);
 
 		/* is this middle part of a bridge? */
@@ -399,7 +420,7 @@ int32 CmdBuildRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 				goto do_clear;
 
 		/* only allow roads pertendicular to bridge */
-		if ((pieces & 5U) == (ti.map5 & 0x01U))
+		if (((pieces & 5U) != 0) == ((ti.map5 & 0x01U) != 0))
 				goto do_clear;
 
 		/* check if clear land under bridge */
@@ -421,17 +442,17 @@ int32 CmdBuildRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 		return cost;
 	} else {
 do_clear:;
-		if (DoCommandByTile(tile, 0, 0, flags & ~DC_EXEC, CMD_LANDSCAPE_CLEAR) == CMD_ERROR)
+		if (CmdFailed(DoCommandByTile(tile, 0, 0, flags & ~DC_EXEC, CMD_LANDSCAPE_CLEAR)))
 			return CMD_ERROR;
 	}
 
 	cost = CheckRoadSlope(ti.tileh, &pieces, existing);
-	if (cost == CMD_ERROR) return_cmd_error(STR_1800_LAND_SLOPED_IN_WRONG_DIRECTION);
+	if (CmdFailed(cost)) return_cmd_error(STR_1800_LAND_SLOPED_IN_WRONG_DIRECTION);
 
-	if (cost && (!_patches.build_on_slopes || (!_patches.ainew_active && _is_ai_player)))
+	if (cost && (!_patches.build_on_slopes || _is_old_ai_player))
 		return CMD_ERROR;
 
-	if (!(ti.type == MP_STREET && (ti.map5 & 0xF0) == 0)) {
+	if (ti.type != MP_STREET || (ti.map5 & 0xF0) != 0) {
 		cost += DoCommandByTile(tile, 0, 0, flags, CMD_LANDSCAPE_CLEAR);
 	} else {
 		// Don't put the pieces that already exist
@@ -449,31 +470,32 @@ do_clear:;
 	if (flags & DC_EXEC) {
 		if (ti.type != MP_STREET) {
 			SetTileType(tile, MP_STREET);
-			_map5[tile] = 0;
-			_map_owner[tile] = _current_player;
+			_m[tile].m5 = 0;
+			_m[tile].m2 = p2;
+			SetTileOwner(tile, _current_player);
 		}
 
-		_map5[tile] |= (byte)pieces;
+		_m[tile].m5 |= pieces;
 
 		MarkTileDirtyByTile(tile);
 	}
 	return cost;
 }
 
-int32 DoConvertStreetRail(uint tile, uint totype, bool exec)
+int32 DoConvertStreetRail(TileIndex tile, uint totype, bool exec)
 {
 	// not a railroad crossing?
-	if ((_map5[tile] & 0xF0) != 0x10) return CMD_ERROR;
+	if (!IsLevelCrossing(tile)) return CMD_ERROR;
 
 	// not owned by me?
 	if (!CheckTileOwnership(tile) || !EnsureNoVehicle(tile)) return CMD_ERROR;
 
 	// tile is already of requested type?
-	if ( (uint)(_map3_hi[tile] & 0xF) == totype) return CMD_ERROR;
+	if (GB(_m[tile].m4, 0, 4) == totype) return CMD_ERROR;
 
 	if (exec) {
 		// change type.
-		_map3_hi[tile] = (_map3_hi[tile] & 0xF0) + totype;
+		SB(_m[tile].m4, 0, 4, totype);
 		MarkTileDirtyByTile(tile);
 	}
 
@@ -481,146 +503,156 @@ int32 DoConvertStreetRail(uint tile, uint totype, bool exec)
 }
 
 
-// Build a long piece of road.
-// x,y = end tile
-// p1 = start tile
-// p2&1 = start tile starts in the 2nd half
-// p2&2 = end tile starts in the 2nd half
-// p2&4 = direction (0 = along x, 1=along y)
+/** Build a long piece of road.
+ * @param x,y end tile of drag
+ * @param p1 start tile of drag
+ * @param p2 various bitstuffed elements
+ * - p2 = (bit 0) - start tile starts in the 2nd half of tile (p2 & 1)
+ * - p2 = (bit 1) - end tile starts in the 2nd half of tile (p2 & 2)
+ * - p2 = (bit 2) - direction: 0 = along x-axis, 1 = along y-axis (p2 & 4)
+ */
 int32 CmdBuildLongRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
-	uint start_tile, end_tile, tile;
-	int mode;
-	int32 cost,ret;
+	TileIndex start_tile, end_tile, tile;
+	int32 cost, ret;
 
 	SET_EXPENSES_TYPE(EXPENSES_CONSTRUCTION);
 
-	start_tile = p1;
-	end_tile = TILE_FROM_XY(x, y);
+	if (p1 >= MapSize()) return CMD_ERROR;
 
-	if (start_tile > end_tile || (start_tile == end_tile && (p2&1))) {
-		uint t = start_tile; start_tile = end_tile; end_tile = t;
+	start_tile = p1;
+	end_tile = TileVirtXY(x, y);
+
+	/* Only drag in X or Y direction dictated by the direction variable */
+	if (!HASBIT(p2, 2) && TileY(start_tile) != TileY(end_tile)) return CMD_ERROR; // x-axis
+	if (HASBIT(p2, 2)  && TileX(start_tile) != TileX(end_tile)) return CMD_ERROR; // y-axis
+
+	/* Swap start and ending tile, also the half-tile drag var (bit 0 and 1) */
+	if (start_tile > end_tile || (start_tile == end_tile && HASBIT(p2, 0))) {
+		TileIndex t = start_tile;
+		start_tile = end_tile;
+		end_tile = t;
 		p2 ^= IS_INT_INSIDE(p2&3, 1, 3) ? 3 : 0;
 	}
 
 	cost = 0;
 	tile = start_tile;
 	// Start tile is the small number.
-	for(;;) {
-		mode = (p2&4) ? 5 : 10;
+	for (;;) {
+		uint bits = HASBIT(p2, 2) ? ROAD_SE | ROAD_NW : ROAD_SW | ROAD_NE;
+		if (tile == end_tile && !HASBIT(p2, 1)) bits &= ROAD_NW | ROAD_NE;
+		if (tile == start_tile && HASBIT(p2, 0)) bits &= ROAD_SE | ROAD_SW;
 
-		if (tile == start_tile && (p2&1))
-			mode &= (4+2);
-		else if (tile == end_tile && !(p2&2))
-			mode &= (1+8);
-
-		ret = DoCommandByTile(tile, mode, 0, flags, CMD_BUILD_ROAD);
-		if (ret == CMD_ERROR) {
-			if (_error_message != STR_1007_ALREADY_BUILT)
-				return CMD_ERROR;
+		ret = DoCommandByTile(tile, bits, 0, flags, CMD_BUILD_ROAD);
+		if (CmdFailed(ret)) {
+			if (_error_message != STR_1007_ALREADY_BUILT) return CMD_ERROR;
 		} else {
 			cost += ret;
 		}
 
-		if (tile == end_tile)
-			break;
+		if (tile == end_tile) break;
 
-		tile += (p2&4)?TILE_XY(0,1):TILE_XY(1,0);
+		tile += HASBIT(p2, 2) ? TileDiffXY(0, 1) : TileDiffXY(1, 0);
 	}
 
-	// already built?
-	if (cost == 0)
-		return CMD_ERROR;
-
-	return cost;
+	return (cost == 0) ? CMD_ERROR : cost;
 }
 
-// Remove a long piece of road.
-// x,y = end tile
-// p1 = start tile
-// p2&1 = start tile starts in the 2nd half
-// p2&2 = end tile starts in the 2nd half
-// p2&4 = direction (0 = along x, 1=along y)
+/** Remove a long piece of road.
+ * @param x,y end tile of drag
+ * @param p1 start tile of drag
+ * @param p2 various bitstuffed elements
+ * - p2 = (bit 0) - start tile starts in the 2nd half of tile (p2 & 1)
+ * - p2 = (bit 1) - end tile starts in the 2nd half of tile (p2 & 2)
+ * - p2 = (bit 2) - direction: 0 = along x-axis, 1 = along y-axis (p2 & 4)
+ */
 int32 CmdRemoveLongRoad(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
-	uint start_tile, end_tile, tile;
-	int32 cost,ret;
+	TileIndex start_tile, end_tile, tile;
+	int32 cost, ret;
 
 	SET_EXPENSES_TYPE(EXPENSES_CONSTRUCTION);
 
-	start_tile = p1;
-	end_tile = TILE_FROM_XY(x, y);
+	if (p1 >= MapSize()) return CMD_ERROR;
 
-	if (start_tile > end_tile || (start_tile == end_tile && (p2&1))) {
-		uint t = start_tile; start_tile = end_tile; end_tile = t;
+	start_tile = p1;
+	end_tile = TileVirtXY(x, y);
+
+	/* Only drag in X or Y direction dictated by the direction variable */
+	if (!HASBIT(p2, 2) && TileY(start_tile) != TileY(end_tile)) return CMD_ERROR; // x-axis
+	if (HASBIT(p2, 2)  && TileX(start_tile) != TileX(end_tile)) return CMD_ERROR; // y-axis
+
+	/* Swap start and ending tile, also the half-tile drag var (bit 0 and 1) */
+	if (start_tile > end_tile || (start_tile == end_tile && HASBIT(p2, 0))) {
+		TileIndex t = start_tile;
+		start_tile = end_tile;
+		end_tile = t;
 		p2 ^= IS_INT_INSIDE(p2&3, 1, 3) ? 3 : 0;
 	}
 
 	cost = 0;
 	tile = start_tile;
 	// Start tile is the small number.
-	for(;;) {
-		uint bits = (p2 & 4) ? ROAD_SE | ROAD_NW : ROAD_SW | ROAD_NE;
-		if (tile == end_tile && !(p2&2)) bits &= ROAD_NW | ROAD_NE;
-		if (tile == start_tile && (p2&1)) bits &= ROAD_SE | ROAD_SW;
+	for (;;) {
+		uint bits = HASBIT(p2, 2) ? ROAD_SE | ROAD_NW : ROAD_SW | ROAD_NE;
+		if (tile == end_tile && !HASBIT(p2, 1)) bits &= ROAD_NW | ROAD_NE;
+		if (tile == start_tile && HASBIT(p2, 0)) bits &= ROAD_SE | ROAD_SW;
 
 		// try to remove the halves.
 		if (bits) {
 			ret = DoCommandByTile(tile, bits, 0, flags, CMD_REMOVE_ROAD);
-			if (ret != CMD_ERROR)
-				cost += ret;
+			if (!CmdFailed(ret)) cost += ret;
 		}
 
-		if (tile == end_tile)
-			break;
+		if (tile == end_tile) break;
 
-		tile += (p2&4)?TILE_XY(0,1):TILE_XY(1,0);
+		tile += HASBIT(p2, 2) ? TileDiffXY(0, 1) : TileDiffXY(1, 0);
 	}
 
-	// already built?
-	if (cost == 0)
-		return CMD_ERROR;
-
-	return cost;
+	return (cost == 0) ? CMD_ERROR : cost;
 }
 
-/* Build a road depot
- * p1 - direction (0-3)
- * p2 - unused
+/** Build a road depot.
+ * @param x,y tile coordinates where the depot will be built
+ * @param p1 depot direction (0 through 3), where 0 is NW, 1 is NE, etc.
+ * @param p2 unused
+ *
+ * @todo When checking for the tile slope,
+ * distingush between "Flat land required" and "land sloped in wrong direction"
  */
-
 int32 CmdBuildRoadDepot(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
 	TileInfo ti;
 	int32 cost;
 	Depot *dep;
-	uint tile;
+	TileIndex tile;
 
 	SET_EXPENSES_TYPE(EXPENSES_CONSTRUCTION);
 
+	if (p1 > 3) return CMD_ERROR; // check direction
+
 	FindLandscapeHeight(&ti, x, y);
 
 	tile = ti.tile;
 
-	if (!EnsureNoVehicle(tile))
-		return CMD_ERROR;
+	if (!EnsureNoVehicle(tile)) return CMD_ERROR;
 
-	if (ti.tileh != 0) {
-		if (!_patches.build_on_slopes || (ti.tileh & 0x10 || !((0x4C >> p1) & ti.tileh) ))
-			return_cmd_error(STR_0007_FLAT_LAND_REQUIRED);
+	if (ti.tileh != 0 && (
+				!_patches.build_on_slopes ||
+				IsSteepTileh(ti.tileh) ||
+				!CanBuildDepotByTileh(p1, ti.tileh)
+			)) {
+		return_cmd_error(STR_0007_FLAT_LAND_REQUIRED);
 	}
 
 	cost = DoCommandByTile(tile, 0, 0, flags, CMD_LANDSCAPE_CLEAR);
-	if (cost == CMD_ERROR)
-		return CMD_ERROR;
+	if (CmdFailed(cost)) return CMD_ERROR;
 
 	dep = AllocateDepot();
-	if (dep == NULL)
-		return CMD_ERROR;
+	if (dep == NULL) return CMD_ERROR;
 
 	if (flags & DC_EXEC) {
-		if (_current_player == _local_player)
-			_last_built_road_depot_tile = (TileIndex)tile;
+		if (IsLocalPlayer()) _last_built_road_depot_tile = tile;
 
 		dep->xy = tile;
 		dep->town_index = ClosestTownFromTile(tile, (uint)-1)->index;
@@ -635,32 +667,30 @@ int32 CmdBuildRoadDepot(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 	return cost + _price.build_road_depot;
 }
 
-static int32 RemoveRoadDepot(uint tile, uint32 flags)
+static int32 RemoveRoadDepot(TileIndex tile, uint32 flags)
 {
 	if (!CheckTileOwnership(tile) && _current_player != OWNER_WATER)
 		return CMD_ERROR;
 
-	if (!EnsureNoVehicle(tile))
-		return CMD_ERROR;
+	if (!EnsureNoVehicle(tile)) return CMD_ERROR;
 
-	if (flags & DC_EXEC) {
-		DoDeleteDepot(tile);
-	}
+	if (flags & DC_EXEC) DoDeleteDepot(tile);
 
 	return _price.remove_road_depot;
 }
 
 #define M(x) (1<<(x))
 
-static int32 ClearTile_Road(uint tile, byte flags) {
+static int32 ClearTile_Road(TileIndex tile, byte flags)
+{
 	int32 ret;
-	byte m5 = _map5[tile];
+	byte m5 = _m[tile].m5;
 
 	if ( (m5 & 0xF0) == 0) {
 		byte b = m5 & 0xF;
 
 		if (! ((1 << b) & (M(1)|M(2)|M(4)|M(8))) ) {
-			if ( (!(flags & DC_AI_BUILDING) || _map_owner[tile]!=OWNER_TOWN) && flags&DC_AUTO)
+			if ((!(flags & DC_AI_BUILDING) || !IsTileOwner(tile, OWNER_TOWN)) && flags & DC_AUTO)
 				return_cmd_error(STR_1801_MUST_REMOVE_ROAD_FIRST);
 		}
 		return DoCommandByTile(tile, b, 0, flags, CMD_REMOVE_ROAD);
@@ -669,8 +699,7 @@ static int32 ClearTile_Road(uint tile, byte flags) {
 			return_cmd_error(STR_1801_MUST_REMOVE_ROAD_FIRST);
 
 		ret = DoCommandByTile(tile, (m5&8)?5:10, 0, flags, CMD_REMOVE_ROAD);
-		if (ret == CMD_ERROR)
-			return CMD_ERROR;
+		if (CmdFailed(ret)) return CMD_ERROR;
 
 		if (flags & DC_EXEC) {
 			DoCommandByTile(tile, 0, 0, flags, CMD_LANDSCAPE_CLEAR);
@@ -702,7 +731,8 @@ typedef struct DrawRoadSeqStruct {
 #include "table/road_land.h"
 
 
-uint GetRoadFoundation(uint tileh, uint bits) {
+uint GetRoadFoundation(uint tileh, uint bits)
+{
 	int i;
 	// normal level sloped building
 	if ((~_valid_tileh_slopes_road[1][tileh] & bits) == 0)
@@ -727,108 +757,121 @@ const byte _road_sloped_sprites[14] = {
 	0,  0
 };
 
+/**
+ * Draw ground sprite and road pieces
+ * @param ti TileInfo
+ * @param road RoadBits to draw
+ * @param ground_type Ground type
+ * @param snow Draw snow
+ * @param flat Draw foundation
+ */
+static void DrawRoadBits(TileInfo *ti, byte road, byte ground_type, bool snow, bool flat)
+{
+	const DrawRoadTileStruct *drts;
+	PalSpriteID image = 0;
+
+	if (ti->tileh != 0) {
+		int foundation;
+		if (flat) {
+			foundation = ti->tileh;
+		} else {
+			foundation = GetRoadFoundation(ti->tileh, road);
+		}
+
+		if (foundation != 0) DrawFoundation(ti, foundation);
+
+		// DrawFoundation() modifies ti.
+		// Default sloped sprites..
+		if (ti->tileh != 0) image = _road_sloped_sprites[ti->tileh - 1] + 0x53F;
+	}
+
+	if (image == 0) image = _road_tile_sprites_1[road];
+
+	if (ground_type == 0) image |= PALETTE_TO_BARE_LAND;
+
+	if (snow) {
+		image += 19;
+	} else if (ground_type > 1 && ground_type != 6) {
+		// Pavement tiles.
+		image -= 19;
+	}
+
+	DrawGroundSprite(image);
+
+	// Return if full detail is disabled, or we are zoomed fully out.
+	if (!(_display_opt & DO_FULL_DETAIL) || _cur_dpi->zoom == 2) return;
+
+	if (ground_type >= 6) {
+		// Road works
+		DrawGroundSprite(HASBIT(road, 3) ? SPR_EXCAVATION_X : SPR_EXCAVATION_Y);
+		return;
+	}
+
+	// Draw extra details.
+	drts = _road_display_table[ground_type][road];
+	while ((image = drts->image) != 0) {
+		int x = ti->x | drts->subcoord_x;
+		int y = ti->y | drts->subcoord_y;
+		byte z = ti->z;
+		if (ti->tileh != 0) z = GetSlopeZ(x, y);
+		AddSortableSpriteToDraw(image, x, y, 2, 2, 0x10, z);
+		drts++;
+	}
+}
+
 static void DrawTile_Road(TileInfo *ti)
 {
-	uint32 image;
+	PalSpriteID image;
 	uint16 m2;
-	const byte *s;
 
 	if ( (ti->map5 & 0xF0) == 0) { // if it is a road the upper 4 bits are 0
-		const DrawRoadTileStruct *drts;
-
-		if (ti->tileh != 0) {
-			int f = GetRoadFoundation(ti->tileh, ti->map5 & 0xF);
-			if (f) DrawFoundation(ti, f);
-
-			// default sloped sprites..
-			if (ti->tileh != 0) {
-				image = _road_sloped_sprites[ti->tileh - 1] + 0x53F;
-			} else  {
-				image = _road_tile_sprites_1[ti->map5 & 0xF];
-			}
-		} else {
-			image = _road_tile_sprites_1[ti->map5 & 0xF];
-		}
-
-		m2 = _map2[ti->tile] & 7;
-
-		if (m2 == 0) image |= 0x3178000;
-
-		if (_map3_hi[ti->tile] & 0x80) {
-			image += 19;
-		} else if (m2 > 1 && m2 != 6) {
-			image -= 19; /* pavement along the road? */
-		}
-
-		DrawGroundSprite(image);
-
-		if (!(_display_opt & DO_FULL_DETAIL) || _cur_dpi->zoom == 2)
-			return;
-
-		if (m2 >= 6) {
-			// roadwork
-			DrawGroundSprite(0x586 + ((ti->map5&8)!=0 ? 0 : 1));
-			return;
-		}
-
-		drts = (const DrawRoadTileStruct*)_road_display_table[m2][ti->map5 & 0xF];
-
-		while ((image = drts->image) != 0) {
-			int x = ti->x | drts->subcoord_x;
-			int y = ti->y | drts->subcoord_y;
-			byte z = ti->z;
-			if (ti->tileh != 0)	z = GetSlopeZ(x, y);
-			AddSortableSpriteToDraw(image, x, y, 2, 2, 0x10, z);
-			drts++;
-		}
+		DrawRoadBits(ti, GB(ti->map5, 0, 4), GB(_m[ti->tile].m4, 4, 3), HASBIT(_m[ti->tile].m4, 7), false);
 	} else if ( (ti->map5 & 0xE0) == 0) { // railroad crossing
 		int f = GetRoadFoundation(ti->tileh, ti->map5 & 0xF);
 		if (f) DrawFoundation(ti, f);
 
-		image = 0x55B;
+		image = GetRailTypeInfo(GB(_m[ti->tile].m4, 0, 4))->base_sprites.crossing;
 
-		if ( (ti->map5 & 8) != 0)
-			image--;
+		if (GB(ti->map5, 3, 1) == 0) image++; /* direction */
 
 		if ( (ti->map5 & 4) != 0)
 			image += 2;
 
-		if ( _map3_hi[ti->tile] & 0x80) {
+		if ( _m[ti->tile].m4 & 0x80) {
 			image += 8;
 		} else {
-			m2 = _map2[ti->tile] & 7;
-			if (m2 == 0) image |= 0x3178000;
+			m2 = GB(_m[ti->tile].m4, 4, 3);
+			if (m2 == 0) image |= PALETTE_TO_BARE_LAND;
 			if (m2 > 1) image += 4;
 		}
 
-		DrawGroundSprite(image + (_map3_hi[ti->tile] & 0xF) * 12);
+		DrawGroundSprite(image);
 	} else {
 		uint32 ormod;
-		int player;
+		PlayerID player;
 		const DrawRoadSeqStruct *drss;
 
 		if (ti->tileh != 0) { DrawFoundation(ti, ti->tileh); }
 
-		ormod = 0x315;
-		player = _map_owner[ti->tile];
+		ormod = PALETTE_TO_GREY;	//was this a bug/problem?
+		player = GetTileOwner(ti->tile);
 		if (player < MAX_PLAYERS)
 			ormod = PLAYER_SPRITE_COLOR(player);
 
-		s = _road_display_datas[ti->map5 & 0xF];
+		drss = _road_display_datas[ti->map5 & 0xF];
 
-		DrawGroundSprite(*(const uint32*)s);
-		s += sizeof(uint32);
-		drss = (const DrawRoadSeqStruct*)s;
+		DrawGroundSprite(drss++->image);
 
-		while ((image=drss->image) != 0) {
-			if (image & 0x8000)
+		for (; drss->image != 0; drss++) {
+			uint32 image = drss->image;
+
+			if (image & PALETTE_MODIFIER_COLOR)
 				image |= ormod;
 			if (_display_opt & DO_TRANS_BUILDINGS) // show transparent depots
-				image = (image & 0x3FFF) | 0x3224000;
+				MAKE_TRANSPARENT(image);
 
 			AddSortableSpriteToDraw(image, ti->x | drss->subcoord_x,
 				ti->y | drss->subcoord_y, drss->width, drss->height, 0x14, ti->z);
-			drss++;
 		}
 	}
 }
@@ -837,30 +880,27 @@ void DrawRoadDepotSprite(int x, int y, int image)
 {
 	uint32 ormod;
 	const DrawRoadSeqStruct *dtss;
-	const byte *t;
 
 	ormod = PLAYER_SPRITE_COLOR(_local_player);
 
-	t = _road_display_datas[image];
+	dtss = _road_display_datas[image];
 
-	x+=33;
-	y+=17;
+	x += 33;
+	y += 17;
 
-	DrawSprite(*(const uint32*)t, x, y);
-	t += sizeof(uint32);
+	DrawSprite(dtss++->image, x, y);
 
-	for(dtss = (const DrawRoadSeqStruct *)t; dtss->image != 0; dtss++) {
+	for(; dtss->image != 0; dtss++) {
 		Point pt = RemapCoords(dtss->subcoord_x, dtss->subcoord_y, 0);
 
 		image = dtss->image;
-		if (image & 0x8000)
-			image |= ormod;
+		if (image & PALETTE_MODIFIER_COLOR) image |= ormod;
 
 		DrawSprite(image, x + pt.x, y + pt.y);
 	}
 }
 
-static uint GetSlopeZ_Road(TileInfo *ti)
+static uint GetSlopeZ_Road(const TileInfo* ti)
 {
 	uint z = ti->z;
 	int th = ti->tileh;
@@ -886,7 +926,7 @@ static uint GetSlopeZ_Road(TileInfo *ti)
 	return z; // normal Z if no slope
 }
 
-static uint GetSlopeTileh_Road(TileInfo *ti)
+static uint GetSlopeTileh_Road(const TileInfo *ti)
 {
 	// check if it's a foundation
 	if (ti->tileh != 0) {
@@ -908,16 +948,14 @@ static uint GetSlopeTileh_Road(TileInfo *ti)
 	return ti->tileh;
 }
 
-static void GetAcceptedCargo_Road(uint tile, AcceptedCargo ac)
+static void GetAcceptedCargo_Road(TileIndex tile, AcceptedCargo ac)
 {
 	/* not used */
 }
 
-static void AnimateTile_Road(uint tile)
+static void AnimateTile_Road(TileIndex tile)
 {
-	if ((_map5[tile] & 0xF0) == 0x10) {
-		MarkTileDirtyByTile(tile);
-	}
+	if (IsLevelCrossing(tile)) MarkTileDirtyByTile(tile);
 }
 
 static const byte _town_road_types[5][2] = {
@@ -937,52 +975,48 @@ static const byte _town_road_types_2[5][2] = {
 };
 
 
-static void TileLoop_Road(uint tile)
+static void TileLoop_Road(TileIndex tile)
 {
 	Town *t;
 	int grp;
 
 	if (_opt.landscape == LT_HILLY) {
 		// Fix snow style if the road is above the snowline
-		if ((_map3_hi[tile] & 0x80) != ((GetTileZ(tile) > _opt.snow_line) ? 0x80 : 0x00)) {
-			_map3_hi[tile] ^= 0x80;
+		if ((_m[tile].m4 & 0x80) != ((GetTileZ(tile) > _opt.snow_line) ? 0x80 : 0x00)) {
+			_m[tile].m4 ^= 0x80;
 			MarkTileDirtyByTile(tile);
 		}
 	} else if (_opt.landscape == LT_DESERT) {
 		// Fix desert style
-		if (GetMapExtraBits(tile) == 1 && !(_map3_hi[tile] & 0x80)) {
-			_map3_hi[tile] |= 0x80;
+		if (GetMapExtraBits(tile) == 1 && !(_m[tile].m4 & 0x80)) {
+			_m[tile].m4 |= 0x80;
 			MarkTileDirtyByTile(tile);
 		}
 	}
 
-	if (_map5[tile] & 0xE0)
+	if (_m[tile].m5 & 0xE0)
 		return;
 
-	if ((_map2[tile] & 7) < 6) {
+	if (GB(_m[tile].m4, 4, 3) < 6) {
 		t = ClosestTownFromTile(tile, (uint)-1);
+
 		grp = 0;
 		if (t != NULL) {
-			// If in the scenario editor, set the owner to a town.
-			if (_game_mode == GM_EDITOR) {
-				_map_owner[tile] = OWNER_TOWN;
-			}
-
 			grp = GetTownRadiusGroup(t, tile);
 
 			// Show an animation to indicate road work
 			if (t->road_build_months != 0 &&
-					!(GetTileDist(t->xy, tile) >= 8 && grp==0) &&
-					(_map5[tile]==5 || _map5[tile]==10)) {
+					!(DistanceManhattan(t->xy, tile) >= 8 && grp == 0) &&
+					(_m[tile].m5==5 || _m[tile].m5==10)) {
 				if (GetTileSlope(tile, NULL) == 0 && EnsureNoVehicle(tile) && CHANCE16(1,20)) {
-					_map2[tile] = ((_map2[tile]&7) <= 1) ? 6 : 7;
+					SB(_m[tile].m4, 4, 3, (GB(_m[tile].m4, 4, 3) <= 1 ? 6 : 7));
 
 					SndPlayTileFx(SND_21_JACKHAMMER, tile);
 					CreateEffectVehicleAbove(
 						TileX(tile) * 16 + 7,
 						TileY(tile) * 16 + 7,
 						0,
-						EV_ROADWORK);
+						EV_BULLDOZER);
 					MarkTileDirtyByTile(tile);
 					return;
 				}
@@ -991,7 +1025,7 @@ static void TileLoop_Road(uint tile)
 
 		{
 			const byte *p = (_opt.landscape == LT_CANDY) ? _town_road_types_2[grp] : _town_road_types[grp];
-			byte b = _map2[tile] & 7;
+			byte b = GB(_m[tile].m4, 4, 3);
 
 			if (b == p[0])
 				return;
@@ -1003,48 +1037,52 @@ static void TileLoop_Road(uint tile)
 			} else {
 				b = 0;
 			}
-			_map2[tile] = (_map2[tile] & ~7) | b;
+			SB(_m[tile].m4, 4, 3, b);
 			MarkTileDirtyByTile(tile);
 		}
 	} else {
 		// Handle road work
+		//XXX undocumented
 
-		uint16 b = _map2[tile];
-		if (b < 0x80) {
-			_map2[tile] = b + 8;
+		byte b = _m[tile].m4;
+		//roadworks take place only
+		//keep roadworks running for 16 loops
+		//lower 4 bits of map3_hi store the counter now
+		if ((b & 0xF) != 0xF) {
+			_m[tile].m4 = b + 1;
 			return;
 		}
-		_map2[tile] = ((b&7) == 6) ? 1 : 2;
+		//roadworks finished
+		_m[tile].m4 = (GB(b, 4, 3) == 6 ? 1 : 2) << 4;
 		MarkTileDirtyByTile(tile);
 	}
 }
 
-void ShowRoadDepotWindow(uint tile);
+void ShowRoadDepotWindow(TileIndex tile);
 
-static void ClickTile_Road(uint tile)
+static void ClickTile_Road(TileIndex tile)
 {
-	if ((_map5[tile] & 0xF0) == 0x20) {
-		ShowRoadDepotWindow(tile);
-	}
+	if (GB(_m[tile].m5, 4, 4) == 2) ShowRoadDepotWindow(tile);
 }
 
 static const byte _road_trackbits[16] = {
 	0x0, 0x0, 0x0, 0x10, 0x0, 0x2, 0x8, 0x1A, 0x0, 0x4, 0x1, 0x15, 0x20, 0x26, 0x29, 0x3F,
 };
 
-static uint32 GetTileTrackStatus_Road(uint tile, TransportType mode)	{
+static uint32 GetTileTrackStatus_Road(TileIndex tile, TransportType mode)
+{
 	if (mode == TRANSPORT_RAIL) {
-		if ((_map5[tile] & 0xF0) != 0x10)
+		if (!IsLevelCrossing(tile))
 			return 0;
-		return _map5[tile] & 8 ? 0x101 : 0x202;
+		return _m[tile].m5 & 8 ? 0x101 : 0x202;
 	} else if  (mode == TRANSPORT_ROAD) {
-		byte b = _map5[tile];
+		byte b = _m[tile].m5;
 		if ((b & 0xF0) == 0) {
 			/* Ordinary road */
-			if (!_road_special_gettrackstatus && (_map2[tile]&7) >= 6)
+			if (!_road_special_gettrackstatus && GB(_m[tile].m4, 4, 3) >= 6)
 				return 0;
 			return _road_trackbits[b&0xF] * 0x101;
-		} else if ((b&0xF0) == 0x10) {
+		} else if (IsLevelCrossing(tile)) {
 			/* Crossing */
 			uint32 r = 0x101;
 			if (b&8) r <<= 1;
@@ -1072,31 +1110,31 @@ static const StringID _road_tile_strings[] = {
 	STR_1814_ROAD,
 };
 
-static void GetTileDesc_Road(uint tile, TileDesc *td)
+static void GetTileDesc_Road(TileIndex tile, TileDesc *td)
 {
-	int i = (_map5[tile] >> 4);
+	int i = (_m[tile].m5 >> 4);
 	if (i == 0)
-		i = (_map2[tile] & 7) + 3;
+		i = GB(_m[tile].m4, 4, 3) + 3;
 	td->str = _road_tile_strings[i - 1];
-	td->owner = _map_owner[tile];
+	td->owner = GetTileOwner(tile);
 }
 
 static const byte _roadveh_enter_depot_unk0[4] = {
 	8, 9, 0, 1
 };
 
-static uint32 VehicleEnter_Road(Vehicle *v, uint tile, int x, int y)
+static uint32 VehicleEnter_Road(Vehicle *v, TileIndex tile, int x, int y)
 {
-	if ((_map5[tile] & 0xF0) == 0x10)	{
-		if (v->type == VEH_Train && (_map5[tile] & 4) == 0) {
+	if (IsLevelCrossing(tile)) {
+		if (v->type == VEH_Train && GB(_m[tile].m5, 2, 1) == 0) {
 			/* train crossing a road */
 			SndPlayVehicleFx(SND_0E_LEVEL_CROSSING, v);
-			_map5[tile] |= 4;
+			SB(_m[tile].m5, 2, 1, 1);
 			MarkTileDirtyByTile(tile);
 		}
-	} else if ((_map5[tile]&0xF0) ==  0x20){
+	} else if (GB(_m[tile].m5, 4, 4) == 2) {
 		if (v->type == VEH_Road && v->u.road.frame == 11) {
-			if (_roadveh_enter_depot_unk0[_map5[tile]&3] == v->u.road.state) {
+			if (_roadveh_enter_depot_unk0[GB(_m[tile].m5, 0, 2)] == v->u.road.state) {
 				RoadVehEnterDepot(v);
 				return 4;
 			}
@@ -1105,37 +1143,34 @@ static uint32 VehicleEnter_Road(Vehicle *v, uint tile, int x, int y)
 	return 0;
 }
 
-static void VehicleLeave_Road(Vehicle *v, uint tile, int x, int y)
+static void VehicleLeave_Road(Vehicle *v, TileIndex tile, int x, int y)
 {
-	if ((_map5[tile] & 0xF0) == 0x10 && v->type == VEH_Train && v->next == NULL) {
-		_map5[tile] &= ~4;
+	if (IsLevelCrossing(tile) && v->type == VEH_Train && v->next == NULL) {
+		// Turn off level crossing lights
+		SB(_m[tile].m5, 2, 1, 0);
 		MarkTileDirtyByTile(tile);
 	}
 }
 
-static void ChangeTileOwner_Road(uint tile, byte old_player, byte new_player)
+static void ChangeTileOwner_Road(TileIndex tile, PlayerID old_player, PlayerID new_player)
 {
-	byte b;
-
 	// road/rail crossing where the road is owned by the current player?
-	if (old_player == _map3_lo[tile] && (_map5[tile]&0xF0) == 0x10) {
-		_map3_lo[tile] = (new_player == 0xFF) ? OWNER_NONE : new_player;
+	if (old_player == _m[tile].m3 && IsLevelCrossing(tile)) {
+		_m[tile].m3 = (new_player == OWNER_SPECTATOR) ? OWNER_NONE : new_player;
 	}
 
-	if (_map_owner[tile] != old_player)
-		return;
+	if (!IsTileOwner(tile, old_player)) return;
 
-	if (new_player != 255) {
-		_map_owner[tile] = new_player;
+	if (new_player != OWNER_SPECTATOR) {
+		SetTileOwner(tile, new_player);
 	}	else {
-		b = _map5[tile]&0xF0;
-		if (b == 0) {
-			_map_owner[tile] = OWNER_NONE;
-		} else if (b == 0x10) {
-			_map5[tile] = (_map5[tile]&8) ? 0x5 : 0xA;
-			_map_owner[tile] = _map3_lo[tile];
-			_map3_lo[tile] = 0;
-			_map3_hi[tile] &= 0x80;
+		if (GB(_m[tile].m5, 4, 4) == 0) {
+			SetTileOwner(tile, OWNER_NONE);
+		} else if (IsLevelCrossing(tile)) {
+			_m[tile].m5 = (_m[tile].m5&8) ? 0x5 : 0xA;
+			SetTileOwner(tile, _m[tile].m3);
+			_m[tile].m3 = 0;
+			_m[tile].m4 &= 0x80;
 		} else {
 			DoCommandByTile(tile, 0, 0, DC_EXEC, CMD_LANDSCAPE_CLEAR);
 		}
